@@ -62,9 +62,13 @@ import collections
 import functools
 import itertools
 import contextlib
+import warnings
 
 import six
-from irc.jaraco import always_iterable
+from jaraco.itertools import always_iterable, infinite_call
+from jaraco.functools import Throttler
+from jaraco.stream import buffer
+from more_itertools.recipes import consume
 
 try:
     import pkg_resources
@@ -74,10 +78,10 @@ except ImportError:
 from . import connection
 from . import events
 from . import functools as irc_functools
-from . import buffer
-from . import schedule
 from . import features
 from . import ctcp
+from . import message
+from . import schedule
 
 log = logging.getLogger(__name__)
 
@@ -105,6 +109,7 @@ class PrioritizedHandler(
         "when sorting prioritized handlers, only use the priority"
         return self.priority < other.priority
 
+
 class Reactor(object):
     """
     Processes events from one or more IRC server connections.
@@ -121,8 +126,8 @@ class Reactor(object):
     corresponding connection.
 
     The methods of most interest for an IRC client writer are server,
-    add_global_handler, remove_global_handler, execute_at,
-    execute_delayed, execute_every, process_once, and process_forever.
+    add_global_handler, remove_global_handler,
+    process_once, and process_forever.
 
     This is functionally an event-loop which can either use it's own
     internal polling loop, or tie into an external event-loop, by
@@ -151,11 +156,12 @@ class Reactor(object):
     are guarded by a mutex.
     """
 
+    scheduler_class = schedule.DefaultScheduler
+
     def __do_nothing(*args, **kwargs):
         pass
 
-    def __init__(self, on_connect=__do_nothing, on_disconnect=__do_nothing,
-            on_schedule=__do_nothing):
+    def __init__(self, on_connect=__do_nothing, on_disconnect=__do_nothing):
         """Constructor for Reactor objects.
 
         on_connect: optional callback invoked when a new connection
@@ -164,13 +170,7 @@ class Reactor(object):
         on_disconnect: optional callback invoked when a socket is
         disconnected.
 
-        on_schedule: optional callback, usually supplied by an external
-        event loop, to indicate in float seconds that the client needs to
-        process events that many seconds in the future. An external event
-        loop will implement this callback to schedule a call to
-        process_timeout.
-
-        The three arguments mainly exist to be able to use an external
+        The arguments mainly exist to be able to use an external
         main loop (for example Tkinter's or PyGTK's main app loop)
         instead of calling the process_forever method.
 
@@ -180,11 +180,13 @@ class Reactor(object):
 
         self._on_connect = on_connect
         self._on_disconnect = on_disconnect
-        self._on_schedule = on_schedule
+
+        scheduler = self.scheduler_class()
+        assert isinstance(scheduler, schedule.IScheduler)
+        self.scheduler = scheduler
 
         self.connections = []
         self.handlers = {}
-        self.delayed_commands = []  # list of DelayedCommands
         # Modifications to these shared lists and dict need to be thread-safe
         self.mutex = threading.RLock()
 
@@ -219,14 +221,7 @@ class Reactor(object):
         See documentation for Reactor.__init__.
         """
         with self.mutex:
-            while self.delayed_commands:
-                command = self.delayed_commands[0]
-                if not command.due():
-                    break
-                command.function()
-                if isinstance(command, schedule.PeriodicCommand):
-                    self._schedule_command(command.next())
-                del self.delayed_commands[0]
+            self.scheduler.run_pending()
 
     @property
     def sockets(self):
@@ -272,8 +267,8 @@ class Reactor(object):
         # Otherwise no other thread would ever be able to change
         # the shared state of a Reactor object running this function.
         log.debug("process_forever(timeout=%s)", timeout)
-        while 1:
-            self.process_once(timeout)
+        one = functools.partial(self.process_once, timeout=timeout)
+        consume(infinite_call(one))
 
     def disconnect_all(self, message=""):
         """Disconnects all connections."""
@@ -330,13 +325,16 @@ class Reactor(object):
 
         Arguments:
 
-            at -- Execute at this time (standard "time_t" time).
+            at -- Execute at this time (a standard Unix timestamp).
             function -- Function to call.
             arguments -- Arguments to give the function.
         """
+        warnings.warn(
+            "Call Reactor.scheduler.execute_every directly",
+            DeprecationWarning,
+        )
         function = functools.partial(function, *arguments)
-        command = schedule.DelayedCommand.at_time(at, function)
-        self._schedule_command(command)
+        self.scheduler.execute_at(at, function)
 
     def execute_delayed(self, delay, function, arguments=()):
         """
@@ -346,9 +344,12 @@ class Reactor(object):
         function -- Function to call.
         arguments -- Arguments to give the function.
         """
+        warnings.warn(
+            "Call Reactor.scheduler.execute_every directly",
+            DeprecationWarning,
+        )
         function = functools.partial(function, *arguments)
-        command = schedule.DelayedCommand.after(delay, function)
-        self._schedule_command(command)
+        self.scheduler.execute_after(delay, function)
 
     def execute_every(self, period, function, arguments=()):
         """
@@ -358,14 +359,12 @@ class Reactor(object):
         function -- Function to call.
         arguments -- Arguments to give the function.
         """
+        warnings.warn(
+            "Call Reactor.scheduler.execute_every directly",
+            DeprecationWarning,
+        )
         function = functools.partial(function, *arguments)
-        command = schedule.PeriodicCommand.after(period, function)
-        self._schedule_command(command)
-
-    def _schedule_command(self, command):
-        with self.mutex:
-            bisect.insort(self.delayed_commands, command)
-            self._on_schedule(command.delay.total_seconds())
+        self.scheduler.execute_every(period, function)
 
     def dcc(self, dcctype="chat"):
         """Creates and returns a DCCConnection object.
@@ -403,14 +402,15 @@ class Reactor(object):
             self.connections.remove(connection)
             self._on_disconnect(connection.socket)
 
-_cmd_pat = "^(:(?P<prefix>[^ ]+) +)?(?P<command>[^ ]+)( *(?P<argument> .+))?"
+_cmd_pat = "^(@(?P<tags>[^ ]*) )?(:(?P<prefix>[^ ]+) +)?(?P<command>[^ ]+)( *(?P<argument> .+))?"
 _rfc_1459_command_regexp = re.compile(_cmd_pat)
 
+
+@six.add_metaclass(abc.ABCMeta)
 class Connection(object):
     """
     Base class for IRC connections.
     """
-    __metaclass__ = abc.ABCMeta
 
     @abc.abstractproperty
     def socket(self):
@@ -420,7 +420,7 @@ class Connection(object):
         self.reactor = reactor
 
     ##############################
-    ### Convenience wrappers.
+    ### Convenience wrappers - deprecated; do not use
 
     def execute_at(self, at, function, arguments=()):
         self.reactor.execute_at(at, function, arguments)
@@ -431,6 +431,7 @@ class Connection(object):
     def execute_every(self, period, function, arguments=()):
         self.reactor.execute_every(period, function, arguments)
 
+
 class ServerConnectionError(IRCError):
     pass
 
@@ -532,19 +533,15 @@ class ServerConnection(Connection):
         This method returns the (real) server name, or, more
         specifically, what the server calls itself.
         """
-
-        if self.real_server_name:
-            return self.real_server_name
-        else:
-            return ""
+        return self.real_server_name or ""
 
     def get_nickname(self):
         """Get the (real) nick name.
 
         This method returns the (real) nickname.  The library keeps
         track of nick changes, so it might not be the nick name that
-        was passed to the connect() method.  """
-
+        was passed to the connect() method.
+        """
         return self.real_nickname
 
     @contextlib.contextmanager
@@ -583,31 +580,19 @@ class ServerConnection(Connection):
             self._process_line(line)
 
     def _process_line(self, line):
-        source = None
-        command = None
-        arguments = None
         event = Event("all_raw_messages", self.get_server_name(), None,
             [line])
         self._handle_event(event)
 
-        m = _rfc_1459_command_regexp.match(line)
-        if m.group("prefix"):
-            prefix = m.group("prefix")
-            if not self.real_server_name:
-                self.real_server_name = prefix
-            source = NickMask(prefix)
+        grp = _rfc_1459_command_regexp.match(line).group
 
-        if m.group("command"):
-            command = m.group("command").lower()
+        source = NickMask.from_group(grp("prefix"))
+        command = self._command_from_group(grp("command"))
+        arguments = message.Arguments.from_group(grp('argument'))
+        tags = message.Tag.from_group(grp('tags'))
 
-        if m.group("argument"):
-            a = m.group("argument").split(" :", 1)
-            arguments = a[0].split()
-            if len(a) == 2:
-                arguments.append(a[1])
-
-        # Translate numerics into more readable strings.
-        command = events.numeric.get(command, command)
+        if source and not self.real_server_name:
+            self.real_server_name = source
 
         if command == "nick":
             if source.nick == self.real_nickname:
@@ -619,58 +604,67 @@ class ServerConnection(Connection):
         elif command == "featurelist":
             self.features.load(arguments)
 
-        if command in ["privmsg", "notice"]:
-            target, message = arguments[0], arguments[1]
-            messages = ctcp.dequote(message)
+        handler = (
+            self._handle_message
+            if command in ["privmsg", "notice"]
+            else self._handle_other
+        )
+        handler(arguments, command, source, tags)
 
-            if command == "privmsg":
-                if is_channel(target):
-                    command = "pubmsg"
-            else:
-                if is_channel(target):
-                    command = "pubnotice"
-                else:
-                    command = "privnotice"
-
-            for m in messages:
-                if isinstance(m, tuple):
-                    if command in ["privmsg", "pubmsg"]:
-                        command = "ctcp"
-                    else:
-                        command = "ctcpreply"
-
-                    m = list(m)
-                    log.debug("command: %s, source: %s, target: %s, "
-                        "arguments: %s", command, source, target, m)
-                    event = Event(command, source, target, m)
-                    self._handle_event(event)
-                    if command == "ctcp" and m[0] == "ACTION":
-                        event = Event("action", source, target, m[1:])
-                        self._handle_event(event)
-                else:
-                    log.debug("command: %s, source: %s, target: %s, "
-                        "arguments: %s", command, source, target, [m])
-                    event = Event(command, source, target, [m])
-                    self._handle_event(event)
+    def _handle_message(self, arguments, command, source, tags):
+        target, msg = arguments[:2]
+        messages = ctcp.dequote(msg)
+        if command == "privmsg":
+            if is_channel(target):
+                command = "pubmsg"
         else:
-            target = None
-
-            if command == "quit":
-                arguments = [arguments[0]]
-            elif command == "ping":
-                target = arguments[0]
+            if is_channel(target):
+                command = "pubnotice"
             else:
-                target = arguments[0]
-                arguments = arguments[1:]
+                command = "privnotice"
+        for m in messages:
+            if isinstance(m, tuple):
+                if command in ["privmsg", "pubmsg"]:
+                    command = "ctcp"
+                else:
+                    command = "ctcpreply"
 
-            if command == "mode":
-                if not is_channel(target):
-                    command = "umode"
+                m = list(m)
+                log.debug("command: %s, source: %s, target: %s, "
+                          "arguments: %s, tags: %s", command, source, target, m, tags)
+                event = Event(command, source, target, m, tags)
+                self._handle_event(event)
+                if command == "ctcp" and m[0] == "ACTION":
+                    event = Event("action", source, target, m[1:], tags)
+                    self._handle_event(event)
+            else:
+                log.debug("command: %s, source: %s, target: %s, "
+                          "arguments: %s, tags: %s", command, source, target, [m], tags)
+                event = Event(command, source, target, [m], tags)
+                self._handle_event(event)
 
-            log.debug("command: %s, source: %s, target: %s, "
-                "arguments: %s", command, source, target, arguments)
-            event = Event(command, source, target, arguments)
-            self._handle_event(event)
+    def _handle_other(self, arguments, command, source, tags):
+        target = None
+        if command == "quit":
+            arguments = [arguments[0]]
+        elif command == "ping":
+            target = arguments[0]
+        else:
+            target = arguments[0] if arguments else None
+            arguments = arguments[1:]
+        if command == "mode":
+            if not is_channel(target):
+                command = "umode"
+        log.debug("command: %s, source: %s, target: %s, "
+                  "arguments: %s, tags: %s", command, source, target, arguments, tags)
+        event = Event(command, source, target, arguments, tags)
+        self._handle_event(event)
+
+    @staticmethod
+    def _command_from_group(group):
+        command = group.lower()
+        # Translate numerics into more readable strings.
+        return events.numeric.get(command, command)
 
     def _handle_event(self, event):
         """[Internal]"""
@@ -901,27 +895,30 @@ class ServerConnection(Connection):
         # unless you've been connected for at least 5 minutes!
         self.send_raw("QUIT" + (message and (" :" + message)))
 
+    def _prep_message(self, string):
+        # The string should not contain any carriage return other than the
+        # one added here.
+        if '\n' in string:
+            msg = "Carriage returns not allowed in privmsg(text)"
+            raise InvalidCharacters(msg)
+        bytes = string.encode('utf-8') + b'\r\n'
+        # According to the RFC http://tools.ietf.org/html/rfc2812#page-6,
+        # clients should not transmit more than 512 bytes.
+        if len(bytes) > 512:
+            msg = "Messages limited to 512 bytes including CR/LF"
+            raise MessageTooLong(msg)
+        return bytes
+
     def send_raw(self, string):
         """Send raw string to the server.
 
         The string will be padded with appropriate CR LF.
         """
-        # The string should not contain any carriage return other than the
-        # one added here.
-        if '\n' in string:
-            raise InvalidCharacters(
-                "Carriage returns not allowed in privmsg(text)")
-        bytes = string.encode('utf-8') + b'\r\n'
-        # According to the RFC http://tools.ietf.org/html/rfc2812#page-6,
-        # clients should not transmit more than 512 bytes.
-        if len(bytes) > 512:
-            raise MessageTooLong(
-                "Messages limited to 512 bytes including CR/LF")
         if self.socket is None:
             raise ServerNotConnectedError("Not connected.")
         sender = getattr(self.socket, 'write', self.socket.send)
         try:
-            sender(bytes)
+            sender(self._prep_message(string))
             log.debug("TO SERVER: %s", string)
         except socket.error:
             # Ouch!
@@ -996,30 +993,7 @@ class ServerConnection(Connection):
         Set a keepalive to occur every ``interval`` on this connection.
         """
         pinger = functools.partial(self.ping, 'keep-alive')
-        self.reactor.execute_every(period=interval, function=pinger)
-
-
-class Throttler(object):
-    """
-    Rate-limit a function (or other callable)
-    """
-    def __init__(self, func, max_rate=float('Inf')):
-        if isinstance(func, Throttler):
-            func = func.func
-        self.func = func
-        self.max_rate = max_rate
-        self.reset()
-
-    def reset(self):
-        self.last_called = 0
-
-    def __call__(self, *args, **kwargs):
-        # ensure at least 1/max_rate seconds from last call
-        elapsed = time.time() - self.last_called
-        must_wait = 1 / self.max_rate - elapsed
-        time.sleep(max(0, must_wait))
-        self.last_called = time.time()
-        return self.func(*args, **kwargs)
+        self.reactor.scheduler.execute_every(period=interval, function=pinger)
 
 
 class DCCConnectionError(IRCError):
@@ -1159,9 +1133,8 @@ class DCCConnection(Connection):
             arguments = [chunk]
             log.debug("command: %s, source: %s, target: %s, arguments: %s",
                 command, prefix, target, arguments)
-            self.reactor._handle_event(
-                self,
-                Event(command, prefix, target, arguments))
+            event = Event(command, prefix, target, arguments)
+            self.reactor._handle_event(self, event)
 
     def privmsg(self, text):
         """
@@ -1198,7 +1171,7 @@ class SimpleIRCClient(object):
     handler methods get two arguments: the connection object (same as
     self.connection) and the event object.
 
-    Functionally, any of the event names in `events.py` my be subscribed
+    Functionally, any of the event names in `events.py` may be subscribed
     to by prefixing them with `on_`, and creating a function of that
     name in the child-class of `SimpleIRCClient`. When the event of
     `event_name` is received, the appropriately named method will be
@@ -1275,8 +1248,13 @@ class SimpleIRCClient(object):
 
 
 class Event(object):
-    "An IRC event."
-    def __init__(self, type, source, target, arguments=None):
+    """
+    An IRC event.
+
+    >>> print(Event('privmsg', '@somebody', '#channel'))
+    type: privmsg, source: @somebody, target: #channel, arguments: [], tags: []
+    """
+    def __init__(self, type, source, target, arguments=None, tags=None):
         """
         Initialize an Event.
 
@@ -1296,6 +1274,20 @@ class Event(object):
         if arguments is None:
             arguments = []
         self.arguments = arguments
+        if tags is None:
+            tags = []
+        self.tags = tags
+
+    def __str__(self):
+        tmpl = (
+            "type: {type}, "
+            "source: {source}, "
+            "target: {target}, "
+            "arguments: {arguments}, "
+            "tags: {tags}"
+        )
+        return tmpl.format(**vars(self))
+
 
 def is_channel(string):
     """Check if a string is a channel name.
@@ -1336,14 +1328,14 @@ class NickMask(six.text_type):
     A nickmask (the source of an Event)
 
     >>> nm = NickMask('pinky!username@example.com')
-    >>> print(nm.nick)
-    pinky
+    >>> nm.nick
+    'pinky'
 
-    >>> print(nm.host)
-    example.com
+    >>> nm.host
+    'example.com'
 
-    >>> print(nm.user)
-    username
+    >>> nm.user
+    'username'
 
     >>> isinstance(nm, six.text_type)
     True
@@ -1358,8 +1350,8 @@ class NickMask(six.text_type):
     Some messages omit the userhost. In that case, None is returned.
 
     >>> nm = NickMask('irc.server.net')
-    >>> print(nm.nick)
-    irc.server.net
+    >>> nm.nick
+    'irc.server.net'
     >>> nm.userhost
     >>> nm.host
     >>> nm.user
@@ -1390,6 +1382,11 @@ class NickMask(six.text_type):
         user, sep, host = userhost.partition('@')
         return user or None
 
+    @classmethod
+    def from_group(cls, group):
+        return cls(group) if group else None
+
+
 def _ping_ponger(connection, event):
     "A global handler for the 'ping' event"
     connection.pong(event.target)
